#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay_basic.h>
#include <avr/eeprom.h>
#include <string.h>


/*********************
 *     RefGen        *
 *  (C) OK1DX 2019   *
 *  ok1dx@amsat.org  *
 *********************
 
 version 1.0
 - first release
 
 Purpose:
 To give precise and clean source of 10 MHz reference signal, that can be used with GPS satellite signal and also without it.
 When the GPS signal is not available, the unit uses stored frequency correction.
 Current software version is frequency synchronized on GPS, but not phase synchronized.
 The frequency deviation should be in range of 10 mHz (when GPS signal available).
 It uses 1 PPS (pulse per second) widely available on GPS units.
 Price for precision is the time constant of control loop, correct sync is reached in minutes and also higher power consumption (OCXO heating).
 
 Note: the quality of OCXO is the key of high quality 10 MHz signal. Not only stability and precision, but especially clean spectrum
 (spectral density of noise close to carrier frequency). TCXO is unable to deliver similar results.
 
 How it works:
 - OCXO is controlled (tuned) from 12bit DAC
 - program calculates DAC value as 16bit, but only 12 MSB are sent to DAC
 - lower 4 bits are used just for smooth calculation (or perhaps later for more precise control)
 - OCXO is a clock source for MCU
 - MCU uses TIMER0 to continuously count clock pulses
 - PPS signal rising edge samples TIMER0 counter using interrupt, sample value is used to calculate frequency deviation
 - the program displays on LCD also some additional information received from GPS module, like coordinates, locator, date and time
 - the unit can be operated completely without LCD display
 - there are 3 LEDs on board, indicating state of synchronisation: PPS signal correct, tune UP and DOWN.
 - the program incorporates startup time (no sync is performed when OCXO is cold) of about 3 minutes
 - the DAC value in synchronised state is saved into EEPROM for further startup as an initial value
 - the EEPROM writing is performed when DAC value is stable enough for certain time.
 - status of writing is indicated on LCD as well: R - read from EEPROM (initial), * - value is correct and already in EEPROM, W - EEPROM was written.
 - the variables of frequency control loop are displayed on LCD as well (DAC value, clock counter difference from last cycle)
 
 Note 2: don't forget to program fuses as first step (esp. low byte is important - clock source). Additionally you need to program
 not only flash (*.HEX) but also EEPROM (*.EEP)!
 
 */

// Clock 10 MHz external (from OCXO)
// fuse bits:
// low byte 11100000 = 0xE0 - CKDIV = 1 (off), CLKOUT = 1 (off), SUT1 = 1, SUT0 = 0, CKSEL0-3 = 0 (external clock)
// high byte 11011100 = 0xDC default (RSTDISBL = 1, DWEN = 1, SPIEN = 0, WDTON = 1, EESAVE = 1, BODLEVEL 4.3V = 100)
// extended byte 11111001 = 0xF9 default (in programmer seen as 0x01 -> just 3 LSB are important)

// LCD interface:
// PC0 - out D4
// PC1 - out D5 
// PC2 - out D6 
// PC3 - out D7
// PC4 - out, RS 0=INSTR, 1=DATA
// PC5 - out, Enable 1=active

#define RS PC4
#define ENABLE PC5

// Programming interface (input, pull up)
// PC6 - RESET
// PB3 - MOSI
// PB4 - MISO
// PB5 - SCK

// SSI interface
// PB2 - SS  out (high when not active)
#define DAC_SYNC PB2
// PB3 - MOSI (data) out
// PB5 - SCK (clock) out

// LED interface (outputs, 1=LED on)
// PD5 out
#define LED1 5
// PD6 out
#define LED2 6
// PD7 out
#define LED3 7

// not used, input+pull up  - PB0, PB1, PB7, PD3, PD4

// initialisation port constants
// port initialisation
// DDR bit 1=output, 0=input
// PORT bit when input 1=pullup 0=high Z

#define _DDRB  0b00101100
#define _PORTB 0b10111111

#define _DDRC  0b00111111
#define _PORTC 0b11000000

#define _DDRD  0b11100010
#define _PORTD 0b00011000

// IRQ variables
uint8_t ppsctr_;  // TIM0 by former pps pulse (used by IRQ only)
uint8_t int0cnt;  // valid pulses 0..3 (used by IRQ only) 

// set by IRQ, read by main loop
volatile uint8_t ppsdif;  // difference between new and last TIM0 value
volatile uint8_t flag;
#define RXPPS 0     // bit 0 - 1pps signal received
#define ERRPPS 1    // bit 1 - pps error (missing or false pulse)

// main loop variables
uint16_t dacValue;  // value to be sent to DAC (calculated)
uint16_t dacValueSaved;  // value identical with that in EEPROM
char eStatus;  // displayed to indicate EEPROM status

#define DAC_TIME 30
#define DACWIN 0x0010
uint8_t dacCounter;  // before writing to EEPROM DAC value must be stable for some time.
uint16_t dacValueX;  // used to evaluate stability of DAC value

uint8_t dacIncr;  // DAC increment per 1Hz difference (16x times more, 0x10 means increment by 1 for every 1Hz difference)
uint8_t viewNr = 0; // view switch (cyclically toggled)

#define RXBUFLEN 80
char rxbuf[RXBUFLEN];  // received buffer - not zero terminated
uint8_t rxcount;  // number of characters in rx buffer
uint16_t runcounter;  // measure time since power up (just indicator, overflow not handled)

#define STARTCOUNT 30
uint8_t startCounter; // time measurement for OCXO to heat up

// all strings zero terminated!
char timestring[9]; // store received time
char datestring[9]; // store received date
char latstring[13];  // latitude string
char longstring[13]; // longitude string
char locator[7];  // locator in maidenhead format

char n2a[7];	// buffer for conversion from number to ASCII. Zero terminated

// EEPROM variables
uint8_t EEMEM welcome[] = "RefGen 1.0 OK1DX";
uint16_t EEMEM eeDacValue = 30276;  // 0x8000;  // DAC middle value, default
uint8_t EEMEM eeDacIncr = 0x05; // DAC increment value, int from EEPROM

void rxgps(void);
void gpsToMaidenhead(void);

/*****************  INTERRUPT HANDLERS ***********************/

// interrupt from 1pps input
// strob the counter0 and set flag
// check time window of pps pulse using TIMER1 value
// restart TIMER1
// inform main loop when new valid value arrived
ISR (INT0_vect)
{
	uint8_t ppsctr = TCNT0;
  
  uint16_t tim1val= TCNT1L;
  tim1val += TCNT1H<<8;
	TCNT1H = 0;  // restart TIM1
	TCNT1L = 0;
  
  // time window - more than 999 msec and less than 1001 msec (OCR1A) and previous calculation finished
  if ((tim1val >= 9756) && !(flag & (1<<RXPPS)))
  {
    if(int0cnt == 0)
    {
      int0cnt++; // 1st pulse do nothing, just start preparation
    }
    else
    {
      if (int0cnt > 2)  // at least 4 valid pulses needed to start calculation
      {
        ppsdif = ppsctr - ppsctr_;  // counter difference
        flag |= (1<<RXPPS);  // give signal "ready for calculation"
      }
      else
      {
        int0cnt++;
      }
      ppsctr_ = ppsctr;     // shift previous value
    }
  }
  else
  {
    int0cnt = 0;  // we start again
    flag |= (1<<ERRPPS);  // give signal "pps error"
  }
}


/****************  LCD FUNCTIONS  *******************/

// internally used by send command and send char; not for direct use!
// RS pin set before, RW always 0
// LCD processing time 40 microsec (400 clocks), whole display 2x16=32 chars = 1.3 msec
// asks for GPS data (to avoid UART overflow)
void lcdSend(uint8_t x)
{	
	PORTC = (PORTC & 0xf0) | ((x >> 4) & 0x0f);// upper nimble
	PORTC |= (1<<ENABLE);	// enable pulse RS=old, EN=1
  _delay_loop_1(3);
	PORTC &= ~(1<<ENABLE);
  _delay_loop_1(3);
	PORTC = (PORTC & 0xF0) | (x & 0x0F);  //lower nibble
	PORTC |= (1<<ENABLE);	// enable pulse
  _delay_loop_1(3);
	PORTC &= ~(1<<ENABLE);
	_delay_loop_1(130);  // delay for LCD processing 40  microsec (400 clock cycles / 3)
  rxgps();
}

// send data to LCD (ASCII code of character to be displayed)
void lcdSendChar(uint8_t cmd)
{
	PORTC = _PORTC | (1<<RS); //RS=1, EN=0
	lcdSend(cmd);
}

// prepare command/address for following writing of data
// LCD line 1: cmd = 0x80 to 0x93, line 2: 0xC0 to 0xD3
void lcdSendCmd(uint8_t cmd)
{
	PORTC = _PORTC; //RS=0, EN=0
	lcdSend(cmd);
}

// sends to LCD zero terminated string
// position must be set in advance
// when string length is less than 'length', the rest up to length is filled with spaces
// set legth to 0 if you need to print just the string.
void lcdSendString(char* sbuf, uint8_t length)
{
  uint8_t ctr = 0;

  while(*sbuf != 0)
  {
    lcdSendChar(*sbuf);
    sbuf++;
    ctr++;
  }
  while(ctr < length)
  {
    lcdSendChar(' '); // fill the rest by spaces = overwrite former text
    ctr++;
  }
}

// sends text from EEPROM to display to current position
void lcdEText(uint8_t* addr)
{
	uint8_t c = 0;
	do
	{
		c = eeprom_read_byte(addr);
		addr++;
		if (c)
			lcdSendChar(c);
	}
	while(c);
}

void lcdinit(void)
{
  _delay_loop_2(62500);	// 1 = 0.4 microsec for 10 MHz clock, 1000 = 0.4 msec
  _delay_loop_2(62500);	// 25 msec
  PORTC = _PORTC |0x03;
  PORTC |= (1<<ENABLE);
  _delay_loop_1(3);
  PORTC &= ~(1<<ENABLE);
  
  _delay_loop_2(25000);	// another 10 msec
  PORTC = _PORTC |0x03;
  PORTC |= (1<<ENABLE);
  _delay_loop_1(3);
  PORTC &= ~(1<<ENABLE);
  
  _delay_loop_2(500);	// another 200 microsec
  PORTC = _PORTC |0x03;
  PORTC |= (1<<ENABLE);
  _delay_loop_1(3);
  PORTC &= ~(1<<ENABLE);
  _delay_loop_1(3);
  
  PORTC = _PORTC |0x02;	// for 4 bit mode
  PORTC |= (1<<ENABLE);
  _delay_loop_1(3);
  PORTC &= ~(1<<ENABLE);	// since now we can use standard lcdSend
  _delay_loop_1(130);
  
  lcdSendCmd(0x28);	// 4 bit, 2 lines, 5x7 mode

  lcdSendCmd(0x08);	// display OFF
  lcdSendCmd(0x0C);	// display ON, no cursor, no blink
  lcdSendCmd(0x06);	// increment mode, no display shift
  lcdSendCmd(0x01);	// display clear
  _delay_loop_2(25000);	// another 10 msec
  
  lcdSendCmd(0x80);	// go to 1st line, 1st position
  lcdEText(welcome);
      
  for (uint8_t i = 0; i<100; i++) // total 2.5 sec
    _delay_loop_2(62500);
}

// conversion of unsigned int (value 0 to 65535) into ASCII string, zero terminated
// fills always the whole buffer (6 characters).
void n2ascii(uint16_t x)
{
  uint8_t i;
	uint16_t base = 10000;
	uint8_t posn = 0;	// buffer position
	for (i=0; i<5; i++)
	{
		uint16_t a = x/base;
		x = x % base;
		base /= 10;
		n2a[posn] = (uint8_t)a + '0';
		posn++;
	}
  n2a[posn] = 0;  // termination
  i = 0;
	while (i<4)
	{
		if ((n2a[i] != '0'))
			break;
		n2a[i] = ' ';	// replace leading zeroes by space
		i++;
	}
}

void lcdChangeView(void)
{
  switch(viewNr)
  {
    case 0:  // numeric coordinates, as received from GPS
      lcdSendCmd(0x80);
      lcdSendString("Lat  ",0);
      lcdSendString(latstring,11);
      lcdSendCmd(0xC0);
      lcdSendString("Lng ",0);
      lcdSendString(longstring,12);
      break;
    case 2:  // UTC time and data, as received from GPS
      lcdSendCmd(0x80);
      lcdSendString("UTC:  ",0);  // length 6
      lcdSendString(timestring, 10);
      lcdSendCmd(0xC0);
      lcdSendString("Date: ",0);  // length 6
      lcdSendString(datestring, 10);
      break;
    case 4: // DAC value, counter difference between following PPS
      lcdSendCmd(0x80);
      lcdSendChar(eStatus);
      lcdSendString(" DAC: ",0);  // length 6
      n2ascii(dacValue);
      lcdSendString(n2a,9);
      lcdSendCmd(0xC0); 
      lcdSendString("PPS dif: ",0);  // length 9
      n2ascii((uint16_t)ppsdif);
      lcdSendString(n2a,7);
      break;
    case 6:  // actual locator and run counter
      lcdSendCmd(0x80);
      gpsToMaidenhead();
      lcdSendString("Loc: ",0);  // length 5
      lcdSendString(locator, 11);
      lcdSendCmd(0xC0);
      lcdSendString("Up time:",0); // length 8
      n2ascii(runcounter++);
      lcdSendString(n2a,8);
      if (startCounter < STARTCOUNT)
      {
        startCounter++;
      }
      dacCounter++;
      break;      
  }
  if (viewNr < 7)
    viewNr++;
  else
    viewNr = 0;
}


/******************* GPS MESSAGE PARSING ***************************/

// processes date or time - copies time from message to time buffer into format DD.MM.YY or HH:MM:SS
// input looks like 220219, output 22.02.19 or 22:02:19
// when rx date empty, copies dummy text
// returns pointer to next item in message
// sbuf = pointer to received message to date/time item
// dbuf = destination buffer
// separator = character used as separator
char* cpydatetime(char* sbuf, char* dbuf, char separator)
{ 
  uint8_t ctr = 0;
  if (*sbuf == ',')  // data empty
  {
    for(ctr=0; ctr<3; ctr++)
    {
      *dbuf = '?';
      dbuf++;
    }
  }
  while((*sbuf != ',') && (*sbuf != 0))
  {
    if (ctr < 6)
    {
      *dbuf = *sbuf;
      dbuf++;
      if ((ctr == 1) || (ctr == 3))
      {
        *dbuf = separator; // add separator
        dbuf++;
      }
      ctr++;
    }
    sbuf++; 
  }
  if (*sbuf == ',')  // jump over the separator
    sbuf++;
  *dbuf = 0; // string terminator
  return sbuf;
}

// copy position from RX buffer into lat/long buffer
// example input: 4908.65635,N,01333.68153,E
// example output: 4908.65635N  01333.68153E
// if message empty, keep last data unchanged
char* cpyposition(char* sbuf, char* dbuf)
{ 
  uint8_t ctr = 0;
  if(*sbuf == 0)  // end of buffer - something's wrong
  {
    return sbuf;
  }
  
  if (*sbuf == ',')
  {
    sbuf++;  // no data yet, jump over separator
  }
  else  // data exists
  {
    while ((ctr < 11) && (*sbuf != ',') && (*sbuf != 0))
    {
      *dbuf = *sbuf;
      dbuf++;
      sbuf++;
      ctr++;
    }
    if (*sbuf== ',') // jump over it
      sbuf++;
    if(*sbuf != 0)
    {
      *dbuf = *sbuf;  // E,W,N,S
      dbuf++;
      sbuf++;
    }
    *dbuf = 0; // string terminator
  }
  
  if (*sbuf == ',')  // jump over the 2nd separator
    sbuf++;

  return sbuf;
}

// read a not used item from the message (jumps over it)
char* cpydummy(char* sbuf)
{
  while ((*sbuf != ',') && (*sbuf != 0))
  {
    sbuf++;
  }
  if (*sbuf == ',')  // jump over the separator
    sbuf++;
  return sbuf;
}

// processing of GPRMC message
// buffer example: $GPRMC,041743.00,A,4908.65348,N,01333.68804,E,1.227,,220219,,,A
// hhmmss.ss, A=valid data, lat, long, speed, course, date, -, -  
void rxgprmc(void)
{
  char* pb = rxbuf + 7;
  pb = cpydatetime(pb, timestring, ':');
  pb = cpydummy(pb);
  pb = cpyposition(pb, latstring);
  pb = cpyposition(pb, longstring);
  pb = cpydummy(pb);
  pb = cpydummy(pb);
  cpydatetime(pb, datestring, '.'); 
}

// processing of GPGLL message
// $GPGLL,4908.65653,N,01333.68099,E,153114.00,A,A
// lat, long, hhmmss.ss, A=valid data, -  
void rxgpgll(void)
{
  char* pb = rxbuf + 7;
  pb = cpyposition(pb, latstring);
  pb = cpyposition(pb, longstring);
  cpydatetime(pb, timestring, ':');
}

/******************   LOCATOR CALCULATION ********************/

// convert longitude or latitude into maidenhead format, fills the buffer
// buffer length = 7 (including \0)
// operates in range -90 to 90 degs (directly latitude, for longitude the values to be divided by 2)
// hemi=false N,E   hemi=true S,W
void maidenhead (char* buf, uint8_t tenDeg, uint8_t deg, uint8_t min, uint8_t hemi)
{
  if (hemi)  // west or south
  {
    *buf=(8-tenDeg) + 'A';
    buf++;
    buf++;
    *buf=(9-deg) + '0';
    buf++;
    buf++;
    *buf=(59-min) + 'A';
  }
  else  // east or north
  {
    *buf=(tenDeg + 9) + 'A';
    buf++;
    buf++;
    *buf=deg + '0';
    buf++;
    buf++;
    *buf=(2 * min)/5 + 'A';
  }
}


// converts coordinate strings into maidenhead string
// example input latstring = 4908.65635N  longstring = 01333.68153E
void gpsToMaidenhead(void)
{
  uint8_t ten, deg, min;
  uint8_t hemi;
  
  if((latstring[0] == '?') || (longstring[0] == '?'))  // data not available
    return;
  
  ten = 10*(longstring[0]-'0') + (longstring[1] - '0');
  deg = longstring[2]-'0' + 10 * (ten & 0x01);
  ten = ten/2;
  min = 10*(longstring[3] - '0') + (longstring[4] - '0') + 60 * (deg & 0x01);
  deg = deg/2;
  min = min/2;
  hemi = longstring[11] == 'W';
  maidenhead(locator, ten, deg, min, hemi);   // E-W
  
  ten = (latstring[0]-'0');
  deg = latstring[1]-'0';
  min = 10*(latstring[2] - '0') + (latstring[3] - '0');
  hemi = latstring[10] == 'S';
  maidenhead(locator + 1, ten, deg, min, hemi);   // N-S
  
  locator[6] = 0; //terminator
}

/******************* OCXO FREQUENCY CONTROL  *********************/

// sends 16 bit data message to DAC - bits 4-15 are sent, bits 0-3 ignored (used for accumulation)
// channel = DAC channel 0..3. We use channel 0 only.
void setDAC(uint8_t channel, uint16_t value)
{
  PORTB &= ~(1<<DAC_SYNC);  // SYNC signal activein low
  // channel number (2bits), power up PD = 1, all regs updated LDAC = 0, MSB 4 bits
  SPDR = ((channel & 0x03) << 6) | 0x20 | ((value >> 12) & 0x0F);
  while (!(SPSR & (1<<SPIF))); // wait till data sent
  uint8_t dummy = SPDR;  // needed to clear flag
  SPDR = (value >> 4) & 0xff;   // LSB
  while (!(SPSR & (1<<SPIF)));
  dummy = SPDR;  // needed to clear flag
  PORTB |= (1<<DAC_SYNC); // SYNC not active
}


// should be called so that collision with PPS interrupt is avoided.
// calculation of PLL correction to DAC value, LED control
// 10 MHz = 1e7 pulses/sec. 1e7 = 0x989680
// counted by 8 bit counter including overflow only low 8 bit available = 0x80.
// so the value between 2 following 1 pps pulses = 0x80 = 128 in case the freq is exact.
// so we are able to detect deviation in range of +/- 128 Hz, that is sufficient
// because OCXO can tune just about +/- 25 Hz (slope 1ppm/V, range 5V)
// OCXO positive slope (higher ADC -> higher OCXO freq)
// LED1 = ADC incremented, LED2 = PPM signal received, LED3 = ADC decremented
void calculatePLL(void)
{
  PORTD &= ~((1<<LED1) | (1<<LED2) | (1<<LED3));  // all LEDs OFF
  
  if (flag & (1<<RXPPS))  // bit set when 3 successive correct 1ppm pulses received
  {
    uint8_t ppsdif_ = ppsdif;
    flag &= ~(1<<RXPPS);
    
    if ((ppsdif_ > 78) && (ppsdif_ < 178)) // accepted  +/- 50 Hz
    {
      PORTD |= (1<<LED2);  // LED 2 on - pulse valid
      uint16_t dacValueOld = dacValue;
      
      if (ppsdif_ > 128)  // OCXO freq too high, tune down
      {
        dacValue -= (ppsdif_ - 128) * dacIncr;
        PORTD |= (1<<LED3);  // LED3 on - ADC decremented
        PORTD &= ~(1<<LED1); // LED1 off
        if (dacValue > dacValueOld)  // detect underflow
        {
          dacValue = 0;  // set DAC to minimum
        }
        setDAC(0, dacValue);
      }
      
      else if (ppsdif_ < 128) // OCXO freq too low, tune up
      {
        dacValue += (128 - ppsdif_) * dacIncr;
        PORTD |= (1<<LED1);  // LED 1 on - ADC incremented
        PORTD &= ~(1<<LED3); // LED3 off
        if (dacValue < dacValueOld)  // detect overflow
        {
          dacValue = 0xFFFF;  // set DAC to maximum
        }
        setDAC(0, dacValue);
      }
      
      else
      {
        PORTD &= ~((1<<LED1) | (1<< LED3)); // LED1 and LED3 off = no ADC change
      }
    }
    else
    {
      flag |= (1<<ERRPPS);  // invalidate pulse
    }
  }
  
  // Value is stable in window
  if ((dacValue > (dacValueX + DACWIN)) || ((dacValue + DACWIN) < dacValueX) || (flag & (1<<ERRPPS)))
  {
    flag &= ~(1<<ERRPPS);
    dacCounter = 0; // out of window, restart timer cycle 
    dacValueX = dacValue;
  }

  // if DAC value stable for long enough time, write it to EEPROM
  if (dacCounter > DAC_TIME)
  {
    dacCounter = 0;
    eStatus = '*';  // is correct, no need to save it to eeprom
    if ((dacValue & 0xFFF0) != (dacValueSaved & 0xFFF0))  // don't overwrite the same number (minimise writing cycles)
    {
      eeprom_write_word(&eeDacValue, dacValue);
      eStatus = 'W';  // written
      dacValueSaved = dacValue;
    }
  }
}


/********************  TOP LEVEL FUNCTIONS  ********************/

// called cyclically - checks whether there are GPS data
// updates msg buffers
// call processing functions for different messages
// call conversion to maidenhead format
// returns immediately (no loops)
void rxgps(void)
{
 while (UCSR0A & (1<<RXC0))
 {
    char x = UDR0; // read one character from USART
    
    if (rxcount == 0) // RX not yet started
    {
      if (x == '$') // $ received
      {
        rxbuf[rxcount++] = x; // put it into buffer and start receiving
      }
    }
    else if ((x == '*') || (x == 0x0D) || (x == 0x0A))
    {
      // message processing
      rxbuf[rxcount] = 0; // terminate
      if (rxcount > 7)  // minimum message length
      {
        if (strncmp(rxbuf,"$GPRMC",6) == 0)
        {
          rxgprmc();
        }
        else if (strncmp(rxbuf,"$GPGLL",6) == 0)
        {
          rxgpgll();
        } 
      }
      rxcount = 0;  // done, continue with next message
    }
    else
    {
      if (rxcount < RXBUFLEN-2)
      {
        rxbuf[rxcount++] = x;  // insert following char into buffer
      }
      else
      {
        rxcount = 0;  // buffer overflow error, start from scratch
      }
    }
  }
}

// initialize all hardware - ports, timer, usart, LCD,...
void initHW(void)
{
  // general data direction: DDRxn = 0 -> input, 1 -> output. PORTxn = 0 -> pullup OFF, 1 -> pullup ON
  
  DDRB = _DDRB;
  PORTB = _PORTB;

  DDRC = _DDRC;
  PORTC = _PORTC;
  
  DDRD = _DDRD;
  PORTD = _PORTD;
    
// TIMER0 INIT
  TCCR0A = 0;  // normal operation
  TCCR0B = 1<<CS00;  // counts system clock directly
  TIMSK0 = 0;  // no interrupt

// TIMER1 INIT
  // runs in CTC 4 mode (WGM10=0, WGM11=0, WGM12=1, WGM13=0). Counter is incremented up to value in reg OCR1A, then zeroed.
  TCCR1A = 0;  // normal operation, no signals connected to pins
  TCCR1B = (1<<CS12) | (1<<CS10) | (1<<WGM12);	// prescaler 1:1024, 1 pulse per 102.4 microsec, CTC mode 4 (up to OCR1A)
  OCR1AH = 0x26;  // 0x262f = 9775 counts = 1001 msec period
  OCR1AL = 0x2F;
  OCR1BH = 0x01;  // 0x01e8 = 488 counts = 50 msec since restart TIM1
  OCR1BL = 0xE8;
  TCNT1H = 0;	// counter value high byte must be written first
  TCNT1L = 0;
  TIFR1 = 0;
  TIMSK1 = 0;	// no interrupt
  
//INT0 INIT
  EICRA = (1<<ISC00) | (1<<ISC01); // rising edge on INT0
  EIMSK = 1<<INT0; // enable INT0 interrupt
  
//USART init
  UCSR0A = 0;  // all standard
  UCSR0B = (1<<RXEN0);  // enable RX, no interrupt
  UBRR0H = 0;
  UBRR0L = 64;  // 9600 Baud by 10 MHz


// SPI
  SPCR = (1<<SPE) | (1<<MSTR) | (1<<CPOL) | (1<<SPR0);  // enable, master, clock mode, clock/16
  SPSR = 0;  // SPI2X0 = 0

  sei();	// global interrupt enable; fun starts here
}


/********************  MAIN FUNCTION ****************/


// main program loop. Performs initialisation and goes into infinite loop, performing
// different tasks. The task flags are set by interrupt procedures
int main (void)
{
  // init global variables
  flag = 0;
  ppsdif = 128;  // difference between new and last TIM0 value
  int0cnt = 0;  // valid pulses 0..3 (used by IRQ only) 
  rxcount = 0;  // number of characters in buffer
  viewNr = 0;
  runcounter = 0;
  startCounter = 0;
  dacCounter = 0;

  // all strings zero terminated!
  timestring[0] = '*';
  timestring[1] = 0;
  datestring[0] = '*';
  datestring[1] = 0;
  latstring[0] = '?';
  latstring[1] = 0;
  longstring[0] = '?';
  longstring[1] = 0;
  locator[0] = '?';
  locator[1] = 0;
  
  dacValue = eeprom_read_word(&eeDacValue);  //power-on DAC value
  dacValueX = dacValueSaved = dacValue;
  dacIncr = eeprom_read_byte(&eeDacIncr);  // initial DAC increment (can be later adjusted!)
  eStatus = 'R';
  
  initHW();
  setDAC(0, dacValue);
  lcdinit();
  
  setDAC(0, dacValue);
  
  while (1)  // endless loop
  {
    rxgps(); // process NMEA data

    if (TIFR1 & 0x04)  // 50 msec after restart of TIM1 - OCFB flag, every 1 sec
    {
      TIFR1 |= 0x04;  // clear the OCFB flag
      lcdChangeView();  // display results - toggling different screen views
      if (startCounter >= STARTCOUNT)  // only when OCXO is heated up
      {
        calculatePLL();  // process PPS measurement results (measured in IRQ)
      }
    }
  }
}
